#include <nds.h>
#include <nds/arm9/console.h>
#include <stdio.h>
#include <stdlib.h>
#include "mathematics.h"
#include "metacube.h"
#include "gfx_engine.h"
#include "global.h"





void MakeRandomShape (struct shape *dest, s8 xsize, s8 ysize, s8 zsize)
{
  u8 x, y, z ;
  u8 count_block = 0 ;
  u8 v ;

  if (xsize <= 0) dest->xsize = (u8) 1 + (rand () % (MAX_SHAPE_DIM - 1)) ;
  else dest->xsize = xsize ;
  if (ysize <= 0) dest->ysize = (u8) 1 + (rand () % (MAX_SHAPE_DIM - 1)) ;
  else dest->ysize = ysize ;
  if (zsize <= 0) dest->zsize = (u8) 1 + (rand () % (MAX_SHAPE_DIM - 1)) ;
  else dest->zsize = zsize ;

  for (x = 0; x < dest->xsize; x++) {
    for (y = 0; y < dest->ysize; y++) {
      for (z = 0; z < dest->zsize; z++) {
	v = (u8) (rand () % 2) ;
	dest->data[x][y][z] = v ;
	if (v != Empty_Cell) count_block ++ ;
      }
    }
  }

  /* Ensure the shape is not empty. */
  if (count_block == 0) {
    x = rand () % dest->xsize ;
    y = rand () % dest->ysize ;
    z = rand () % dest->zsize ;
    dest->data[x][y][z] = Cube_In_Cell ;
    count_block++ ;
  }

  dest->weight = count_block ;
  /* Initially place the shape' center above the metacube's center */
  /* on the X-Z plane, 2 units above the metacube on the Y axis.   */
  /* The shape will then able to translate, but never rotate. The  */
  /* metacube will be able to only rotate.                         */
  dest->xorigin = ((MCUBEDIM - dest->xsize) / 2) ;
  dest->yorigin = MCUBEDIM + 1 ;
  dest->zorigin = ((MCUBEDIM - dest->zsize) / 2) ;
}





enum fitting_status FitP (struct shape *s)
{
  enum fitting_status status = Fit_Inside ;
  u8 x_in_shape, y_in_shape, z_in_shape ;
  s8 world_x, world_y, world_z ;
  /* To prevent too many dereferenciations in loops. */
  s8 xorigin = s->xorigin ;
  s8 yorigin = s->yorigin ;
  s8 zorigin = s->zorigin ;

  for (x_in_shape = 0; x_in_shape < s->xsize; x_in_shape++) {
    for (y_in_shape = 0; y_in_shape < s->ysize; y_in_shape++) {
      for (z_in_shape = 0; z_in_shape < s->zsize; z_in_shape++) {
	if (s->data[x_in_shape][y_in_shape][z_in_shape] != Empty_Cell) {
	  /* This shape's cell contains a block. */
	  world_x = xorigin + ((s8) x_in_shape) ;
	  world_y = yorigin + ((s8) y_in_shape) ;
	  world_z = zorigin + ((s8) z_in_shape) ;
	  if (world_x < 0) return (Fit_Not) ;
	  if (world_x >= MCUBEDIM) return (Fit_Not) ;
	  if (world_z < 0) return (Fit_Not) ;
	  if (world_z >= MCUBEDIM) return (Fit_Not) ;
	  if (world_y < 0) return (Fit_Not) ;
	  /* It may fit, but anyway this will not be a correct position.   */
	  /* because we still are (possibly partially) above the metacube. */
	  if (world_y >= MCUBEDIM) status = Fit_Outside ;
	  else {
	    /* Ensure there is no collision inside the metacube. */
	    /* We are sure we are fully inside the metacube.     */
	    if (global_current_metacube->cells[world_x][world_y][world_z] != 0)
	      return (Fit_Not) ;
	  }
	} /* End of if (s->data[x_in_shape][y_in_shape][z_in_shape] != 0). */
      }   /* End of for (z_in_shape = 0; z_in_shape < s->zsize; ... */
    }     /* End of for (y_in_shape = 0; y_in_shape < s->ysize; ... */
  }       /* End of for (x_in_shape = 0; x_in_shape < s->xsize; ... */
  return (status) ;
}





bool TranslateZForwardShape (struct shape *s)
{
  s->zorigin++ ;
  if ((FitP (s)) == Fit_Outside) return (true) ;
  else {
    s->zorigin-- ;
    return (false) ;
  }
}

bool TranslateZBackwardShape (struct shape *s)
{
  s->zorigin-- ;
  if ((FitP (s))  == Fit_Outside) return (true) ;
  else {
    s->zorigin++ ;
    return (false) ;
  }
}

bool TranslateXLeftShape (struct shape *s)
{
  s->xorigin-- ;
  if ((FitP (s)) == Fit_Outside) return (true) ;
  else {
    s->xorigin++ ;
    return (false) ;
  }
}

bool TranslateXRightShape (struct shape *s)
{
  s->xorigin++ ;
  if ((FitP (s)) == Fit_Outside) return (true) ;
  else {
    s->xorigin-- ;
    return (false) ;
  }
}


s8 DropShapeP (struct shape *s)
{
  s8 y ;
  s8 old_shape_yorigin = s->yorigin ;
  s8 last_fitting = IMPOSSIBLE_DROP_Z ;

  /* Check top-down if there is room. */
  for (y = MCUBEDIM - 1; y > -s->ysize ; y--) {
    s->yorigin = y ;
    switch (FitP (s)) {
    case Fit_Inside:
      last_fitting = y ;
      break ;

    case Fit_Outside:
      /* Go on checking if we can find a fitting INSIDE the metacube. */
      break ;

    case Fit_Not:
      goto dirty_fitting_failure ;
      break ;
    }
  }

 dirty_fitting_failure:
  /* Restore the original shape Z. */
  s->yorigin = old_shape_yorigin ;
  return (last_fitting) ;
}



void MergeShape (struct shape *s)
{
  u8 block ;
  u8 x_in_shape, y_in_shape, z_in_shape ;
  s8 world_x, world_y, world_z ;
  /* To prevent too many dereferenciations in loops. */
  s8 xorigin = s->xorigin ;
  s8 yorigin = s->yorigin ;
  s8 zorigin = s->zorigin ;

  for (x_in_shape = 0; x_in_shape < s->xsize; x_in_shape++) {
    for (y_in_shape = 0; y_in_shape < s->ysize; y_in_shape++) {
      for (z_in_shape = 0; z_in_shape < s->zsize; z_in_shape++) {
	block = s->data[x_in_shape][y_in_shape][z_in_shape] ;
	if (block != Empty_Cell) {
	  /* This shape's cell contains a block. */
	  world_x = xorigin + ((s8) x_in_shape) ;
	  world_y = yorigin + ((s8) y_in_shape) ;
	  world_z = zorigin + ((s8) z_in_shape) ;
	  global_current_metacube->cells[world_x][world_y][world_z] = block ;
	  global_current_metacube->counts_along_x[world_y][world_z]++ ;
	  global_current_metacube->counts_along_y[world_x][world_z]++ ;
	  global_current_metacube->counts_along_z[world_x][world_y]++ ;
	}   /* End of if (block != 0). */
      }  /* End of for (z_in_shape = 0; z_in_shape < s->zsize; ... */
    }    /* End of for (y_in_shape = 0; y_in_shape < s->ysize; ... */
  }      /* End of for (x_in_shape = 0; x_in_shape < s->xsize; ... */
}



/*
  phi positive
  x' = x cos (phi) - y sin (phi)
  y' = x sin (phi) + y cos (phi)

  x' = -y
  y' = x

  Re-center along x:
  x' = (MCUBEDIM -1) - y
  y' = x
  z' = z
*/
void RotateMetacubeZ90CounterClockwise (struct metacube *src,
					struct metacube *dest)
{
  u16 x, y, z ;
  s16 x_, y_ ;

  for (z = 0; z < MCUBEDIM; z++) {
    for (y = 0; y < MCUBEDIM; y++) {
      x_ = (MCUBEDIM - 1) - y ;
      for (x = 0; x < MCUBEDIM; x++) {
	y_ = x ;
	dest->cells[x_][y_][z] = src->cells[x][y][z] ;
	dest->counts_along_x[y_][z] = src->counts_along_y[x][z] ;
	dest->counts_along_y[x_][z] = src->counts_along_x[y][z] ;
	dest->counts_along_z[x_][y_] = src->counts_along_z[x][y] ;
      }
    }
  }
}





/*
  phi negative
  x' = x cos (phi) - y sin (phi)
  y' = x sin (phi) + y cos (phi)

  x' = y
  y' = -x

  Re-center along y:
  x' = y
  y' = (MCUBEDIM - 1) - x
  z' = z
*/
void RotateMetacubeZ90Clockwise (struct metacube *src, struct metacube *dest)
{
  u16 x, y, z ;
  s16 x_, y_ ;

  for (z = 0; z < MCUBEDIM; z++) {
    for (y = 0; y < MCUBEDIM; y++) {
      x_ = y ;
      for (x = 0; x < MCUBEDIM; x++) {
	y_ = (MCUBEDIM - 1) - x ;
	dest->cells[x_][y_][z] = src->cells[x][y][z] ;
	dest->counts_along_x[y_][z] = src->counts_along_y[x][z] ;
	dest->counts_along_y[x_][z] = src->counts_along_x[y][z] ;
	dest->counts_along_z[x_][y_] = src->counts_along_z[x][y] ;
      }
    }
  }
}






/*
  Because z is flipped, this is Clockwise (phi is positive when rotating
  clockwise).
  phi positive
  y' = y cos (phi) - z sin (phi)
  z' = y sin (phi) + z cos (phi)

  y' = -z
  z' = y

  Re-center along y:
  x' = x
  y' = (MCUBEDIM - 1) - z
  z' = y

*/
void RotateMetacubeX90Clockwise (struct metacube *src, struct metacube *dest)
{
  u16 x, y, z ;
  s16 y_, z_ ;
  s16 debx, deby ;

  printf ("Before X90Clock:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", src->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }

  for (x = 0; x < MCUBEDIM; x++) {
    for (z = 0; z < MCUBEDIM; z++) {
      y_ = (MCUBEDIM - 1) - z ;
      for (y = 0; y < MCUBEDIM; y++) {
	z_ = y ;
	dest->cells[x][y_][z_] = src->cells[x][y][z] ;
	dest->counts_along_x[y_][z_] = src->counts_along_x[y][z] ;
	dest->counts_along_y[x][z_] = src->counts_along_z[x][y] ;
	dest->counts_along_z[x][y_] = src->counts_along_y[x][z] ;
      }
    }
  }

  printf ("After X90Clock:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", dest->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }

}



/*
  Because z is flipped, this is CounterClockwise (phi is negative when
  rotating counterclockwise).
  phi negative

  y' = z
  z' = -y

  Re-center along z:
  x' = x
  y' = z
  z' = (MCUBEDIM - 1) - y

*/
void RotateMetacubeX90CounterClockwise (struct metacube *src,
					struct metacube *dest)
{
  u16 x, y, z ;
  s16 y_, z_ ;
  s16 debx, deby ;

  printf ("Before X90Counter:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", src->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }

  for (x = 0; x < MCUBEDIM; x++) {
    for (z = 0; z < MCUBEDIM; z++) {
      y_ = z ;
      for (y = 0; y < MCUBEDIM; y++) {
	z_ = (MCUBEDIM - 1) - y ;
	dest->cells[x][y_][z_] = src->cells[x][y][z] ;
	dest->counts_along_x[y_][z_] = src->counts_along_x[y][z] ;
	dest->counts_along_y[x][z_] = src->counts_along_z[x][y] ;
	dest->counts_along_z[x][y_] = src->counts_along_y[x][z] ;
      }
    }
  }

  printf ("After X90Counter:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", dest->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }
}


/*
  Because x is flipped, this is Clockwise (phi is positive when rotating
  clockwise).
  phi positive

  x' = x cos (phi) + z sin (phi)
  z' = -x sin (phi) + z cos (phi)

  x' = z
  z' = -x

  Re-center along z:
  x' = z
  y' = y
  z' = (MCUBEDIM - 1) - x

*/
void RotateMetacubeY90Clockwise (struct metacube *src, struct metacube *dest)
{
  u16 x, y, z ;
  s16 x_, z_ ;
  s16 debx, deby ;

  printf ("Before Y90Clock:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", src->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }

  for (x = 0; x < MCUBEDIM; x++) {
    for (z = 0; z < MCUBEDIM; z++) {
      x_ = z ;
      for (y = 0; y < MCUBEDIM; y++) {
	z_ = (MCUBEDIM - 1) - x ;
	dest->cells[x_][y][z_] = src->cells[x][y][z] ;
	dest->counts_along_x[y][z_] = src->counts_along_z[x][y] ;
	dest->counts_along_y[x_][z_] = src->counts_along_y[x][z] ;
	dest->counts_along_z[x_][y] = src->counts_along_x[y][z] ;
      }
    }
  }
  printf ("After Y90Clock:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", dest->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }
}



/*
  Because x is flipped, this is CounterClockwise (phi is negative when rotating
  counterclockwise).
  phi negative

  x' = x cos (phi) + z sin (phi)
  z' = -x sin (phi) + z cos (phi)

  x' = -z
  z' = x

  Re-center along  x:
  x' = (MCUBEDIM - 1) - z
  y' = y
  z' = x
*/
void RotateMetacubeY90CounterClockwise (struct metacube *src,
					struct metacube *dest)
{
  u16 x, y, z ;
  s16 x_, z_ ;
  s16 debx, deby ;

  printf ("Before Y90Counter:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", src->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }

  for (x = 0; x < MCUBEDIM; x++) {
    for (z = 0; z < MCUBEDIM; z++) {
      x_ = (MCUBEDIM - 1) - z ;
      for (y = 0; y < MCUBEDIM; y++) {
	z_ = x ;
	dest->cells[x_][y][z_] = src->cells[x][y][z] ;
	dest->counts_along_x[y][z_] = src->counts_along_z[x][y] ;
	dest->counts_along_y[x_][z_] = src->counts_along_y[x][z] ;
	dest->counts_along_z[x_][y] = src->counts_along_x[y][z] ;
      }
    }
  }

  printf ("After Y90Counter:\n") ;
  for (deby = (MCUBEDIM - 1); deby >= 0; deby--) {
    for (debx = 0; debx < MCUBEDIM; debx++)
      printf ("%d ", dest->counts_along_z[debx][deby]) ;
    printf ("\n") ;
  }
}





void MakeRandomMetacube (struct metacube *dest)
{
  u16 x, y, z ;
  u8 v ;

  for (z = 0; z < MCUBEDIM; z++) {
    for (y = 0; y < MCUBEDIM; y++) {
      for (x = 0; x < MCUBEDIM; x++) {
	v = (u8) (rand () % 2) ;
	dest->cells[x][y][z] = v ;
	if (v != Empty_Cell) {
	  dest->counts_along_x[y][z]++ ;
	  dest->counts_along_y[x][z]++ ;
	  dest->counts_along_z[x][y]++ ;
	}
      }
    }
  }
}



void MakeEmptyMetacube (struct metacube *dest)
{
  u16 x, y, z ;

  for (z = 0; z < MCUBEDIM; z++) {
    for (y = 0; y < MCUBEDIM; y++) {
      for (x = 0; x < MCUBEDIM; x++) {
	dest->cells[x][y][z] = Empty_Cell ;
	dest->counts_along_x[y][z] = 0 ;
	dest->counts_along_y[x][z] = 0 ;
	dest->counts_along_z[x][y] = 0 ;
      }
    }
  }
}




/* Assumes that the shape's origins are properly set as in dropped position. */
void CheckDestructions (struct shape *s)
{
  s8 world_start_x, world_start_y, world_start_z ;
  s8 world_stop_x, world_stop_y, world_stop_z ;
  u8 x, y, z ;
  u8 count_lines = 0 ;

  world_start_x = max (0, s->xorigin) ;
  world_start_y = max (0, s->yorigin) ;
  world_start_z = max (0, s->zorigin) ;

  world_stop_x = min (MCUBEDIM, s->xorigin + s->xsize) ;
  world_stop_y = min (MCUBEDIM, s->yorigin + s->ysize) ;
  world_stop_z = min (MCUBEDIM, s->zorigin + s->zsize) ;

  for (y = world_start_y; y < world_stop_y; y++) {
    for (z = world_start_z; z < world_stop_z; z++) {
      if (global_current_metacube->counts_along_x[y][z] == MCUBEDIM) {
	count_lines++ ;
	for (x = 0; x < MCUBEDIM; x++)
	  global_current_metacube->cells[x][y][z] = Deletion_Marked_Cell ;
      }
    }
  }

  for (x = world_start_x; x < world_stop_x; x++) {
    for (z = world_start_z; z < world_stop_z; z++) {
      if (global_current_metacube->counts_along_y[x][z] == MCUBEDIM) {
	count_lines++ ;
	for (y = 0; y < MCUBEDIM; y++)
	  global_current_metacube->cells[x][y][z] = Deletion_Marked_Cell ;
      }
    }
  }

  for (x = world_start_x; x < world_stop_x; x++) {
    for (y = world_start_y; y < world_stop_y; y++) {
      if (global_current_metacube->counts_along_z[x][y] == MCUBEDIM) {
	count_lines++ ;
	for (z = 0; z < MCUBEDIM; z++)
	  global_current_metacube->cells[x][y][z] = Deletion_Marked_Cell ;
      }
    }
  }

  /* Just animate before really erasing the cells if some exist. */
  if (count_lines > 0) ShowDestructions () ;

  /* Scoring for lines: n * (n + 1) / 2. */
  /* In other words: 1 + 2 + 3 + ... + nb_lines. Just factorise the / 2 */
  /* on the SCORE_LINE_DESTROYED constant to allow constant folding.    */
  global_game.score +=
    (count_lines * (count_lines + 1)) * (SCORE_LINE_DESTROYED / 2) ;
  global_game.nb_lines  += count_lines ;
  if (count_lines > 0) {
    for (x = 0; x < MCUBEDIM; x++) {
      for (y = 0; y < MCUBEDIM; y++) {
	for (z = 0; z < MCUBEDIM; z++) {
	  if (global_current_metacube->cells[x][y][z] == Deletion_Marked_Cell){
	    /* Effective deletion. */
	    global_current_metacube->cells[x][y][z] = Empty_Cell ;
	    /* Counters update. */
	    global_current_metacube->counts_along_x[y][z]-- ;
	    global_current_metacube->counts_along_y[x][z]-- ;
	    global_current_metacube->counts_along_z[x][y]-- ;
	  }  /* End of if (global_current_metacube->cells[x][y][z] ==... */
	}   /* End of for (z = 0; z < MCUBEDIM; z++). */
      }    /* End of for (y = 0; y < MCUBEDIM; y++). */
    }      /* End of for (x = 0; x < MCUBEDIM; x++). */
  }      /* End of if (count_lines > 0). */
}
